/***************************************************************************
 *            jackProcess.cpp
 *
 *  Copyright  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "jackProcess.h"
#include <jack/transport.h>
#include "../base/tranches.h"
#include <iostream>
#include <math.h>

using namespace std;

int process(jack_nframes_t nframes, void *arg)
{
	//GET THE PORTS 

	jackProcess* proc = (jackProcess*) arg;
	int nbInputs= proc->inputPorts[0].size();
	int nbOutputs= proc->outputPorts[0].size();

	//we get the input port(s)
	jack_default_audio_sample_t *in[2][nbInputs];
	for(int i=0;i<nbInputs;++i)
	{
		in[0][i]= (jack_default_audio_sample_t *)jack_port_get_buffer (proc->inputPorts[0][i], nframes);	
		in[1][i]= (jack_default_audio_sample_t *)jack_port_get_buffer (proc->inputPorts[1][i], nframes);	
	}
	
	//and the output port(s)
	jack_default_audio_sample_t *out[2][nbOutputs];
	for(int i=0;i< nbOutputs;++i)
	{
		out[0][i] = (jack_default_audio_sample_t *)jack_port_get_buffer (proc->outputPorts[0][i], nframes);	
		out[1][i] = (jack_default_audio_sample_t *)jack_port_get_buffer (proc->outputPorts[1][i], nframes);	

		//put the outputs to 0
		for(unsigned int n=0;n<nframes;++n) //for each sample
		{
			out[0][i][n] = 0 ;	
			out[1][i][n] = 0 ;
		}	
	}

	//GET THE BEAT

	//we get the current Beat and test if the bpm has changed

	jack_position_t pos;
	jack_transport_query(proc->jackClient,&pos);
	int currentBeat = pos.beat-1;
	if(pos.beats_per_minute!=proc->beatsPerMinute && pos.beats_per_minute!=0)
	{
		proc->beatsPerMinute=pos.beats_per_minute;
		proc->setSliceSize(1);	
	}

	//TEST THE REPEAT
	
	//we first test if we should stop
	if(proc->stopSlicingAsked)
	{
		proc->clearSlice();
	}

	//we then test if repeat has been pressed and if a new beat started : to wait for a new beat before starting repeating
	//if(proc->slicingAsked && currentBeat!=proc->previousBeat)
	if(proc->slicingAsked)
	{
		proc->slicing = true;
		proc->slicingAsked = false;
		proc->sliceRecorded = false;
	}
	
	int previousOutput = proc->previousOutput+nbInputs;
	int sliceOutput = proc->sliceOutput+nbInputs;
	int sliceInput = proc->sliceInput;


	//PROCESS

	//if we are in slicing mode 
	if(proc->slicing)
	{
		if(sliceInput==-1)	//slice all the inputs //NEED TO FADE ON ALL THE OUTPUTS ?
		{
			for(unsigned int n=0;n<nframes;++n) //for each sample
			{
				//if recording the slice is not over
				if(!(proc->sliceRecorded))
				{
					//we add all the mix of all theinputs to the slice
					jack_default_audio_sample_t in1=0;
					jack_default_audio_sample_t in2=0;
					for(int i=0;i<nbInputs;++i)
					{
						in1+=in[0][i][n];
						in2+=in[1][i][n];
				
					}
					proc->addSliceSample(in1,in2);
					
					//and we play the sample in the current output
					out[0][sliceOutput][n]=in1;	
					out[1][sliceOutput][n]=in2;	
				}
				else	//if it is over
				{
					//we get the sample from the slice
					jack_default_audio_sample_t s1 = proc->getNextSliceSample(1);
					jack_default_audio_sample_t s2 = proc->getNextSliceSample(2);
					
					//and we play the sample in the current output
					out[0][sliceOutput][n]=s1;	
					out[1][sliceOutput][n]=s2;	
				}
			}
		}
		else		//if we slice only one input
		{

			//we test if the output has changed, to crossfade the sound between the previous and current outputs
			if(previousOutput != sliceOutput)
			{
				//we first fill the slice output 
				for(unsigned int n=0;n<nframes;++n) //for each sample
				{
					//if recording the slice is not over
					if(!(proc->sliceRecorded))
					{
						//we add the sample to the slice
						proc->addSliceSample(in[0][sliceInput][n],in[1][sliceInput][n]);
						
						//and we play the sample in the previous and current output with the fade 
						out[0][previousOutput][n]=in[0][sliceInput][n]*(jack_default_audio_sample_t)(1.0 - (double)n/(double)nframes);	
						out[1][previousOutput][n]=in[1][sliceInput][n]*(jack_default_audio_sample_t)(1.0 - (double)n/(double)nframes);	
						
						out[0][sliceOutput][n]=in[0][sliceInput][n]*(jack_default_audio_sample_t)((double)n/(double)nframes);	
						out[1][sliceOutput][n]=in[1][sliceInput][n]*(jack_default_audio_sample_t)((double)n/(double)nframes);	
						
					}
					else	//if it is over
					{
						//we get the sample from the slice
						jack_default_audio_sample_t s1 = proc->getNextSliceSample(1);
						jack_default_audio_sample_t s2 = proc->getNextSliceSample(2);
						
						//and we play the sample in the previous and current output with the fade 
						out[0][previousOutput][n]=s1*(jack_default_audio_sample_t)(1.0 - (double)n/(double)nframes);	
						out[1][previousOutput][n]=s2*(jack_default_audio_sample_t)(1.0 - (double)n/(double)nframes);	
						
						out[0][sliceOutput][n]=s1*(jack_default_audio_sample_t)((double)n/(double)nframes);	
						out[1][sliceOutput][n]=s2*(jack_default_audio_sample_t)((double)n/(double)nframes);	
					}
				}
				
				//then we fill the outputs for the dry inputs
				for(int i=0;i<nbInputs;++i)
				{
					if(i!=sliceInput && i!=previousOutput)
					{
						for(unsigned int n=0;n<nframes;++n) //for each sample
						{
								out[0][i][n]=in[0][i][n];	
								out[1][i][n]=in[1][i][n];	
						}
					}
				}
			}
			else
			{
			
				//we first fill the slice output
				for(unsigned int n=0;n<nframes;++n) //for each sample
				{
					//if recording the slice is not over
					if(!(proc->sliceRecorded))
					{
						//we add the sample to the slice
						proc->addSliceSample(in[0][sliceInput][n],in[1][sliceInput][n]);
						
						//and we play the sample 
						out[0][sliceOutput][n]=in[0][sliceInput][n];	
						out[1][sliceOutput][n]=in[1][sliceInput][n];	
						
					}
					else	//if it is over
					{
						//we get the sample from the slice
						jack_default_audio_sample_t s1 = proc->getNextSliceSample(1);
						jack_default_audio_sample_t s2 = proc->getNextSliceSample(2);
						
						//we play the sound 
						out[0][sliceOutput][n]=s1;	
						out[1][sliceOutput][n]=s2;	
							
					}
				}
				
				//then we fill the outputs for the dry inputs
				for(int i=0;i<nbInputs;++i)
				{
					if(i!=sliceInput)
					{	
						for(unsigned int n=0;n<nframes;++n) //for each sample
						{
								out[0][i][n]=in[0][i][n];	
								out[1][i][n]=in[1][i][n];	
						}
					}
				}
				
			}
		}
		proc->previousOutput = proc->sliceOutput;
	}
	else	//normal mode	
	{
		//we fill all the outputs
		

		for(int i=0;i<nbInputs;++i)
		{
			for(unsigned int n=0;n<nframes;++n) //for each sample
			{
					out[0][i][n]=in[0][i][n];	
					out[1][i][n]=in[1][i][n];	
			}
		}

		proc->previousOutput = 0;
	}
		
	proc->previousBeat=currentBeat;
	
	
	return 0;
}

int sRate (jack_nframes_t nframes, void *arg)
{
	
	jackProcess* proc = (jackProcess*) arg;
	proc->sampleRate = nframes;
	return 0;
}

void shutdown(void *arg)
{
	jackProcess* proc = (jackProcess*) arg;
	proc->tra->showMessage(true,"Jack ... killed us");
	proc->jackClient=NULL;
}


jackProcess::jackProcess(tranches* t):tra(t),sliceCounter(0),sliceSize(2048),maxSliceSize(2048),slicingAsked(false)
,sliceRecorded(false),slicing(false),sliceInput(0),previousBeat(0),previousOutput(0),fadingTab(NULL)
{
	sliceOutput = tra->getInputNb();
	beatsPerMinute = 120;
	sampleRate = 44100;
}

jackProcess::~jackProcess()
{
	jack_deactivate(jackClient); 
	for(int i=0;i<inputPorts[0].size();++i)
	{
		jack_port_unregister (jackClient,inputPorts[0][i]);
		jack_port_unregister (jackClient,inputPorts[1][i]);
	}
	inputPorts[0].clear();
	inputPorts[1].clear();
	for(int i=0;i<outputPorts[0].size();++i)
	{
		jack_port_unregister (jackClient,outputPorts[0][i]);
		jack_port_unregister (jackClient,outputPorts[1][i]);
	}
	outputPorts[0].clear();
	outputPorts[1].clear();
	jack_client_close(jackClient);
}


int jackProcess::init()
{

	//jack client creation
	jackClient=NULL;
	jackClient = jack_client_open("Tranches",JackUseExactName,NULL);		
	if(!jackClient)
	{
		return -1;
	}

	sampleRate=jack_get_sample_rate(jackClient);

	//sample rate callback 
	jack_set_sample_rate_callback (jackClient, sRate, this);

	//process callback
	jack_set_process_callback (jackClient, process, this);

	//on shutdown

	jack_on_shutdown(jackClient,shutdown,this);	

	//we get the current BPM
	jack_position_t pos;
	jack_transport_query(jackClient,&pos);
	if(pos.beats_per_minute!=0)
	{
		beatsPerMinute = pos.beats_per_minute;
	}

	return 0;
}

int jackProcess::addPort(int inout,std::string st)
{
	if(!inout)//input
	{
		outputPorts[0].insert(outputPorts[0].begin()+inputPorts[0].size(),jack_port_register (jackClient, ("dry-"+st+"-L").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
		outputPorts[1].insert(outputPorts[1].begin()+inputPorts[0].size(),jack_port_register (jackClient, ("dry-"+st+"-R").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
		inputPorts[0].push_back(jack_port_register (jackClient, (st+"-L").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0));
		inputPorts[1].push_back(jack_port_register (jackClient, (st+"-R").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0));
	}
	else//output
	{
		outputPorts[0].push_back(jack_port_register (jackClient, ("wet-"+st+"-L").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
		outputPorts[1].push_back(jack_port_register (jackClient, ("wet-"+st+"-R").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
	}
}

int jackProcess::removePort(int inout,int num)
{
	if(!inout)//input
	{
		jack_port_unregister (jackClient,inputPorts[0][num]);
		jack_port_unregister (jackClient,inputPorts[1][num]);
		inputPorts[0].erase(inputPorts[0].begin()+num);	
		inputPorts[1].erase(inputPorts[1].begin()+num);	
		jack_port_unregister (jackClient,outputPorts[0][num]);
		jack_port_unregister (jackClient,outputPorts[1][num]);
		outputPorts[0].erase(outputPorts[0].begin()+num);	
		outputPorts[1].erase(outputPorts[1].begin()+num);	
	}
	else//output
	{
		jack_port_unregister (jackClient,outputPorts[0][num+inputPorts[0].size()]);
		jack_port_unregister (jackClient,outputPorts[1][num+inputPorts[0].size()]);
		outputPorts[0].erase(outputPorts[0].begin()+num+inputPorts[0].size());	
		outputPorts[1].erase(outputPorts[1].begin()+num+inputPorts[0].size());	
	}
}

int jackProcess::renamePort(int inout,int num,std::string st)
{	
	if(!inout)//input
	{
		jack_port_set_name(inputPorts[0][num],(st+"-L").c_str());	
		jack_port_set_name(inputPorts[1][num],(st+"-R").c_str());
		jack_port_set_name(outputPorts[0][num],("dry-"+st+"-L").c_str());	
		jack_port_set_name(outputPorts[1][num],("dry-"+st+"-R").c_str());	
	}
	else//output
	{
		jack_port_set_name(outputPorts[0][num+inputPorts[0].size()],("wet-"+st+"-L").c_str());	
		jack_port_set_name(outputPorts[1][num+inputPorts[0].size()],("wet-"+st+"-R").c_str());
	}
}

int jackProcess::start()
{
	jack_deactivate(jackClient); 

	if(inputPorts[0].size()>0) //if ports already created
	{
		//delete everything
		for(int i=0;i<inputPorts[0].size();++i)
		{
			jack_port_unregister (jackClient,inputPorts[0][i]);
			jack_port_unregister (jackClient,inputPorts[1][i]);
		}
		inputPorts[0].clear();
		inputPorts[1].clear();
		for(int i=0;i<outputPorts[0].size();++i)
		{
			jack_port_unregister (jackClient,outputPorts[0][i]);
			jack_port_unregister (jackClient,outputPorts[1][i]);
		}
		outputPorts[0].clear();
		outputPorts[1].clear();
	}

	//create the input ports
	for(int i=0;i<tra->getInputNb();++i)
	{
		std::string st = tra->getInputName(i);
		inputPorts[0].push_back(jack_port_register (jackClient, (st+"-L").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0));
		inputPorts[1].push_back(jack_port_register (jackClient, (st+"-R").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0));
	}
	
	//we create the output ports
	
	//the dry outputs
	for(int i=0;i<inputPorts[0].size();++i)
	{
			std::string st = tra->getInputName(i);
			outputPorts[0].push_back(jack_port_register (jackClient, ("dry-"+st+"-L").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
			outputPorts[1].push_back(jack_port_register (jackClient, ("dry-"+st+"-R").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
	}
	//the sliced outputs
	for(int i=0;i<tra->getOutputNb();++i)
	{
			std::string st = tra->getOutputName(i);
			outputPorts[0].push_back(jack_port_register (jackClient, ("wet-"+st+"-L").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
			outputPorts[1].push_back(jack_port_register (jackClient, ("wet-"+st+"-R").c_str(), JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0));
	}

	//activate the client
	if (jack_activate (jackClient)) 
	{
		return -1;
	}

	//set the reverse
	reverse=false;

	//set the slicing Size to one beat, create the tab
	setSliceSize(1);
	return 0 ;
}



jack_client_t* jackProcess::getClient()
{
	return jackClient;
}

int jackProcess::getSampleRate()
{
	return sampleRate;
}

void jackProcess::addSliceSample(jack_default_audio_sample_t in1 ,jack_default_audio_sample_t in2 )
{
	//we add the sample
	sliceSamplesL.push_back(in1);
	sliceSamplesR.push_back(in2);
	//test with the size of the repeat
	//if size reached, slice is recorded
	if(sliceSamplesL.size() >= (unsigned int)sliceSize)
	{
		maxSliceSize=sliceSize;
		sliceRecorded=true;
		sliceCounter=0;
	}
}		

jack_default_audio_sample_t jackProcess::getNextSliceSample(int channel)
{
	//return the sample faded in/out to avoid clicks
	jack_default_audio_sample_t res;
	if(channel==1)
		res = sliceSamplesL[sliceCounter] * fadingTab[sliceCounter];
	else
	{
		res = sliceSamplesR[sliceCounter] * fadingTab[sliceCounter];
		//counter ++
		if(reverse)
			sliceCounter=(sliceCounter<=0)? (sliceSize-1) : sliceCounter-1;
		else
			sliceCounter=(sliceCounter>=(sliceSize-1))? 0 : sliceCounter+1;
	}
	return res;
}


void jackProcess::setSliceSize(double nbBeats)
{	
	int fadeTime=100;

	//we compute the new sliceSize
	int nsize = (int)floor(nbBeats / (beatsPerMinute/60.0) * (double)sampleRate);

	//we change the sliceSize if possible	
	if(slicing)
	{
		if(sliceRecorded)
		{
			sliceSize=(nsize<=maxSliceSize)? nsize : maxSliceSize;
		
			//and we refill the tab
			for(int i=0;i<fadeTime;++i)
			{
				fadingTab[i]=fadingTab[sliceSize-1-i]= double(i)/double(fadeTime);
			}	 
			for(int i=fadeTime;i<=sliceSize-fadeTime;++i)
			{
				fadingTab[i]=1.0;	
			}
		}
	}
	else
	{	
		sliceSize=nsize;

		//we rebuild the fading tab
		if(fadingTab)
			delete(fadingTab);
		fadingTab = new jack_default_audio_sample_t[sliceSize];

		//and we fill it
		for(int i=0;i<fadeTime;++i)
		{
			fadingTab[i]=fadingTab[sliceSize-1-i]= double(i)/double(fadeTime);
		}	 
		for(int i=fadeTime;i<=sliceSize-fadeTime;++i)
		{
			fadingTab[i]=1.0;	
		}
	}
	
	//we put the sliceCounter at the right place	
	sliceCounter = sliceCounter % sliceSize;
}

void jackProcess::setSliceInput(int in)
{
	sliceInput=in;
}

void jackProcess::setSliceOutput(int out)
{
	sliceOutput=out;
}

void jackProcess::startSlicing()
{
	slicingAsked = true;	
}

void jackProcess::stopSlicing()
{
	stopSlicingAsked=true;
}

void jackProcess::toggleReverse(int val)
{
	reverse=(val==0)? false: true;
}

void jackProcess::clearSlice()
{
	slicingAsked=false;
	stopSlicingAsked=false;
	slicing=false;
	sliceSamplesL.clear();	
	sliceSamplesR.clear();	
}

