/***************************************************************************
 *            jackProcess.h
 *
 *  Copyright  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _JACKPROCESS_H
#define _JACKPROCESS_H

#include <iostream>
#include <jack/jack.h>
#include <jack/ringbuffer.h>
#include <jack/types.h>
#include <cstdlib>
#include <vector>

class tranches;

class jackProcess
{
	public:
		jackProcess(tranches*);
		~jackProcess();
		int init();
		int start();
		int addPort(int,std::string);
		int removePort(int,int);
		int renamePort(int,int,std::string);
		jack_client_t* getClient();
		int getSampleRate();
		void setSliceSize(double);
		void setSliceInput(int);
		void setSliceOutput(int);
		void startSlicing();
		void stopSlicing();
		void toggleReverse(int);
		void addSliceSample(jack_default_audio_sample_t ,jack_default_audio_sample_t);		
		jack_default_audio_sample_t getNextSliceSample(int);
		void clearSlice();
		friend int process(jack_nframes_t , void *);
		friend int sRate(jack_nframes_t , void *);
		friend void shutdown(void*);
	
	protected:
		tranches* tra;
		jack_client_t *jackClient;
		jack_nframes_t sampleRate;	
		std::vector<jack_port_t*> outputPorts[2];
		std::vector<jack_port_t*> inputPorts[2];
		double beatsPerMinute;

		std::vector<jack_default_audio_sample_t> sliceSamplesL;		 	
		std::vector<jack_default_audio_sample_t> sliceSamplesR;		 	
		int sliceCounter;	
		int sliceSize;	
		int maxSliceSize;	
		bool slicingAsked;
		bool stopSlicingAsked;
		bool sliceRecorded;	
		bool slicing;
		int sliceOutput;
		int sliceInput;
		bool reverse;
		jack_default_audio_sample_t *fadingTab;

		int previousBeat;
		int previousOutput;
			
};


#endif 
