/***************************************************************************
 *          traParser.cpp
 *
 *  Copyright  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "traParser.h"
#include "tranches.h"
#include <iostream>
#include <cstdio>
#include <cstdlib>
using namespace std;

traParser::traParser(tranches *t):tra(t)
{
	LIBXML_TEST_VERSION
	doc=NULL;
	root_element=NULL;
}

traParser::~traParser()
{
	xmlCleanupParser();
}


int traParser::createTree(char* f)
{
	fileName=f;

   	doc= xmlReadFile(f,NULL,0);
    	root_element = xmlDocGetRootElement(doc);
	if (doc == NULL || root_element==NULL ) 
	{
		std::string f(fileName);
        	tra->showMessage(true,"Failed to parse "+f);
		return -1;
    	}

	if (xmlStrcmp(root_element->name, (const xmlChar *) "tranches")) 
	{
		tra->showMessage(true,"Document of the wrong type, root node != tranches");
		xmlFreeDoc(doc);
		return -1;
	}

	return 0;
}


int traParser::parseTree()
{
	int res=0;
	char* val=NULL;

	//we start with the first node
	xmlNodePtr a_node = root_element->children;

	//get tranches file version
	val=(char*)xmlGetProp(root_element,(xmlChar*)"version");
	if(val==NULL)
	{
		if(parse005(a_node))
			res = -1;
		else
			res = saveToFile(fileName);
	}
	else
		res = parseCurrent(a_node);

    	xmlFreeDoc(doc);
	return res;
}


int traParser::parse005(xmlNode *a_node)
{
    	xmlNodePtr cur_node = NULL;
	
	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
 	{
		char* value=NULL;    
        	if (cur_node->type == XML_ELEMENT_NODE) 
		{
			switch(cur_node->name[0])
			{
				case 's':	//slice node
					{
						//toggle midi Channel
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"toggle_midi_channel");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : toggle_midi_channel missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(0,0,atoi(value));
						value=NULL;						

						//toggle midi CC or note
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"toggle_midi_cc_or_note");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : toggle_midi_cc_or_note missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(0,1,atoi(value));
						value=NULL;						
						
						//frequency midi Channel
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"frequency_midi_channel");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : frequency_midi_channel missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(1,0,atoi(value));
						value=NULL;

						//frequency midi CC 
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"frequency_midi_cc");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : frequency_midi_cc missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(1,1,atoi(value));
						value=NULL;

						//input midi Channel
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"input_midi_channel");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : input_midi_channel missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(2,0,atoi(value));
						value=NULL;
						
						//input midi CC 
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"input_midi_cc");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : input_midi_cc missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(2,1,atoi(value));
						value=NULL;

						//output midi Channel
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"output_midi_channel");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : output_midi_channel missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(3,0,atoi(value));
						value=NULL;
						
						//output midi CC 
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"output_midi_cc");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : output_midi_cc missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(3,1,atoi(value));
						value=NULL;

						//reverse midi Channel
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"reverse_midi_channel");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : reverse_midi_channel missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(4,0,atoi(value));
						value=NULL;						

						//reverse midi CC or note
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"reverse_midi_cc_or_note");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : reverse_midi_cc_or_note missing in the slice node"); 
							return -1;
						}
						tra->setMidiControl(4,1,atoi(value));
						value=NULL;						
					}
					break;
				case 'o':	//output node
					{
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"name");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : name missing in an output node"); 
							return -1;
						}
						tra->addOutput(value);
						value=NULL;
					}
					break;
				case 'i':	//input node
					{
						value = (char*)xmlGetProp(cur_node,(xmlChar*)"name");
						if(value==NULL)
						{
							tra->showMessage(true,"Problem in the .tra file : name missing in an input node"); 
							return -1;
						}
						tra->addInput(value,0,0);
						value=NULL;
					}
					break;
				default : // problem
					{
						tra->showMessage(true,"Argh, problem in the .tra file !!!");
						tra->showMessage(true,"Unknown Node "+std::string((char*)(cur_node->name)));
						return -1;
					}
					break;
			}


        	}
    	}
	return 0;
}


int traParser::parseCurrent(xmlNodePtr a_node)
{
    	xmlNodePtr cur_node = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
 	{
		char* val = NULL;    	
		std::string nodeName((const char*)(cur_node->name));
        	if (cur_node->type == XML_ELEMENT_NODE) 
		{
			if(nodeName=="inputs")
			{
				if(parseInputs(cur_node))
					return -1;
			}
			else if(nodeName=="outputs")
			{
				if(parseOutputs(cur_node))
					return -1;
			}
			else if(nodeName=="slice")
			{
				if(parseSlice(cur_node))
					return -1;
			}
			else if(nodeName=="repeat")
			{
				if(parseRepeat(cur_node))
					return -1;
			}
			else if(nodeName=="swap")
			{
				if(parseSwap(cur_node))
					return -1;
			}
			else if(nodeName=="granulize")
			{
				if(parseGranulize(cur_node))
					return -1;
			}
		}
    	}
	return 0;
}

int traParser::parseInputs(xmlNodePtr cur_node)
{
	char* value=NULL;
	unsigned short channel = -1;
	unsigned short cc = -1;
    	xmlNodePtr cNode = NULL;

	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"midi_channel");
	if(value==NULL)
	{
		tra->showMessage(false,"Problem in the .tra file : midi_channel missing in the inputs node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : midi_cc missing in the inputs node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(0,0,channel);
	tra->setMidiControl(0,1,cc);
	
	for (cNode = cur_node->children; cNode; cNode = cNode->next) 
 	{
		value = NULL;    	
		std::string nodeName((const char*)(cNode->name));
        	if (cNode->type == XML_ELEMENT_NODE) 
		{
			if(nodeName=="input")
			{
				channel=-1;
				cc=-1;

				//name
				value = (char*)xmlGetProp(cNode,(xmlChar*)"name");
				if(value==NULL)
				{
					tra->showMessage(true,"Problem in the .tra file : name missing in an input node"); 
					return -1;
				}
				std::string name(value);
				value=NULL;

				//midi Channel
				value = (char*)xmlGetProp(cNode,(xmlChar*)"midi_channel");
				if(value==NULL)
				{
					tra->showMessage(true,"Problem in the .tra file : midi_channel missing in an input node"); 
				}
				else
				{
					channel=(unsigned short)(atoi(value));
				}
				value=NULL;						

				//midi CC or note
				value = (char*)xmlGetProp(cNode,(xmlChar*)"midi_cc");
				if(value==NULL)
				{
					tra->showMessage(true,"Problem in the .tra file : midi_cc missing in an input node"); 
				}
				else
				{
					cc=(unsigned short)(atoi(value));
				}
				
				tra->addInput(name,channel,cc);
			}
		}
	}
	
	return 0;
}

int traParser::parseOutputs(xmlNodePtr cur_node)
{
	char* value=NULL;
	unsigned short channel = -1;
	unsigned short cc = -1;
    	xmlNodePtr cNode = NULL;

	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : midi_channel missing in the outputs node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : midi_cc missing in the outputs node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(1,0,channel);
	tra->setMidiControl(1,1,cc);
	
	for (cNode = cur_node->children; cNode; cNode = cNode->next) 
 	{
		value = NULL;    	
		std::string nodeName((const char*)(cNode->name));
        	if (cNode->type == XML_ELEMENT_NODE) 
		{
			if(nodeName=="output")
			{
				//name
				value = (char*)xmlGetProp(cNode,(xmlChar*)"name");
				if(value==NULL)
				{
					tra->showMessage(true,"Problem in the .tra file : name missing in an output node"); 
					return -1;
				}
				std::string name(value);
				tra->addOutput(name);
			}
		}
	}
	
	return 0;
}

int traParser::parseSlice(xmlNodePtr cur_node)
{
	char* value=NULL;
	unsigned short channel = -1;
	unsigned short cc = -1;
    	xmlNodePtr cNode = NULL;

	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : midi_channel missing in the slice node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : midi_cc missing in the slice node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(2,0,channel);
	tra->setMidiControl(2,1,cc);

	return 0;
}

int traParser::parseSwap(xmlNodePtr cur_node)
{
	char* value=NULL;
	unsigned short channel = -1;
	unsigned short cc = -1;
    	xmlNodePtr cNode = NULL;

	//SWAP
	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"swap_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : swap_midi_channel missing in the swap node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"swap_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : swap_midi_cc missing in the swap node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(3,0,channel);
	tra->setMidiControl(3,1,cc);

	value=NULL;
	channel = -1;
	cc = -1;

	//SLICE 2
	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"slice1_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : slice1_midi_channel missing in the swap node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"slice1_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : slice1_midi_cc missing in the swap node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(4,0,channel);
	tra->setMidiControl(4,1,cc);

	//SLICE 2
	value=NULL;
	channel = -1;
	cc = -1;

	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"slice2_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : slice2_midi_channel missing in the swap node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"slice2_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : slice2_midi_cc missing in the swap node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(5,0,channel);
	tra->setMidiControl(5,1,cc);

	return 0;
}

int traParser::parseRepeat(xmlNodePtr cur_node)
{
	char* value=NULL;
	unsigned short channel = -1;
	unsigned short cc = -1;
    	xmlNodePtr cNode = NULL;

	//REPEAT
	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"repeat_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : repeat_midi_channel missing in the repeat node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"repeat_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : repeat_midi_cc missing in the repeat node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(6,0,channel);
	tra->setMidiControl(6,1,cc);

	value=NULL;
	channel = -1;
	cc = -1;

	//REVERSE
	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"reverse_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : reverse_midi_channel missing in the repeat node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"reverse_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : reverse_midi_cc missing in the repeat node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(7,0,channel);
	tra->setMidiControl(7,1,cc);

	return 0;
}

int traParser::parseGranulize(xmlNodePtr cur_node)
{
	char* value=NULL;
	unsigned short channel = -1;
	unsigned short cc = -1;
    	xmlNodePtr cNode = NULL;

	//GRANULIZE
	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"granulize_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : granulize_midi_channel missing in the granulize node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"granulize_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : granulize_midi_cc missing in the granulize node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(8,0,channel);
	tra->setMidiControl(8,1,cc);

	value=NULL;
	channel = -1;
	cc = -1;

	//GRAIN SIZE
	//midi Channel
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"size_midi_channel");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : size_midi_channel missing in the granulize node"); 
	}
	else
	{
		channel=(unsigned short)(atoi(value));
	}
	value=NULL;						

	//midi CC or note
	value = (char*)xmlGetProp(cur_node,(xmlChar*)"size_midi_cc");
	if(value==NULL)
	{
		tra->showMessage(true,"Problem in the .tra file : size_midi_cc missing in the granulize node"); 
	}
	else
	{
		cc=(unsigned short)(atoi(value));
	}

	tra->setMidiControl(9,0,channel);
	tra->setMidiControl(9,1,cc);
	
	return 0;
}

int traParser::saveToFile(char* filename)
{
	char buf[32];
	
	xmlDocPtr doc = NULL;  
    	xmlNodePtr rootNode = NULL, node1 = NULL,node2=NULL;

	doc = xmlNewDoc(BAD_CAST "1.0");
    	rootNode = xmlNewNode(NULL, BAD_CAST "tranches");
	xmlNewProp(rootNode, BAD_CAST "version", BAD_CAST TRANCHES_VERSION );
    	xmlDocSetRootElement(doc, rootNode);
	
	//input nodes
	node1 = xmlNewChild(rootNode, NULL, BAD_CAST "inputs",NULL);
	sprintf(buf, "%d", tra->getMidiControl(0,0));
	xmlNewProp(node1, BAD_CAST "midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(0,1));
	xmlNewProp(node1, BAD_CAST "midi_cc", BAD_CAST buf);


	//for each input
	for(int i=0;i<tra->getInputNb();++i)
	{
		node2 = xmlNewChild(node1, NULL, BAD_CAST "input",NULL);
		xmlNewProp(node2, BAD_CAST "name", BAD_CAST tra->getInputName(i).c_str());
		sprintf(buf, "%d", tra->getMidiControl(i+NB_MIDI_CONTROLS,0));
		xmlNewProp(node2, BAD_CAST "midi_channel", BAD_CAST buf);
		sprintf(buf, "%d", tra->getMidiControl(i+NB_MIDI_CONTROLS,1));
		xmlNewProp(node2, BAD_CAST "midi_cc", BAD_CAST buf);
	}

	//outputs nodes	
	node1 = xmlNewChild(rootNode, NULL, BAD_CAST "outputs",NULL);
	sprintf(buf, "%d", tra->getMidiControl(1,0));
	xmlNewProp(node1, BAD_CAST "midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(1,1));
	xmlNewProp(node1, BAD_CAST "midi_cc", BAD_CAST buf);
	for(int i=0;i<tra->getOutputNb();++i)
	{
		node2 = xmlNewChild(node1, NULL, BAD_CAST "output",NULL);
		xmlNewProp(node2, BAD_CAST "name", BAD_CAST tra->getOutputName(i).c_str() );
	}
	

	//slice node
	node1 = xmlNewChild(rootNode, NULL, BAD_CAST "slice",NULL);
	sprintf(buf, "%d", tra->getMidiControl(2,0));
	xmlNewProp(node1, BAD_CAST "midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(2,1));
	xmlNewProp(node1, BAD_CAST "midi_cc", BAD_CAST buf);
	
	//swap node
	node1 = xmlNewChild(rootNode, NULL, BAD_CAST "swap",NULL);
	sprintf(buf, "%d", tra->getMidiControl(3,0));
	xmlNewProp(node1, BAD_CAST "swap_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(3,1));
	xmlNewProp(node1, BAD_CAST "swap_midi_cc", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(4,0));
	xmlNewProp(node1, BAD_CAST "slice1_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(4,1));
	xmlNewProp(node1, BAD_CAST "slice1_midi_cc", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(5,0));
	xmlNewProp(node1, BAD_CAST "slice2_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(5,1));
	xmlNewProp(node1, BAD_CAST "slice2_midi_cc", BAD_CAST buf);
	
	//repeat node
	node1 = xmlNewChild(rootNode, NULL, BAD_CAST "repeat",NULL);
	sprintf(buf, "%d", tra->getMidiControl(6,0));
	xmlNewProp(node1, BAD_CAST "repeat_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(6,1));
	xmlNewProp(node1, BAD_CAST "repeat_midi_cc", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(7,0));
	xmlNewProp(node1, BAD_CAST "reverse_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(7,1));
	xmlNewProp(node1, BAD_CAST "reverse_midi_cc", BAD_CAST buf);

	//granulize node
	node1 = xmlNewChild(rootNode, NULL, BAD_CAST "granulize",NULL);
	sprintf(buf, "%d", tra->getMidiControl(8,0));
	xmlNewProp(node1, BAD_CAST "granulize_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(8,1));
	xmlNewProp(node1, BAD_CAST "granulize_midi_cc", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(9,0));
	xmlNewProp(node1, BAD_CAST "size_midi_channel", BAD_CAST buf);
	sprintf(buf, "%d", tra->getMidiControl(9,1));
	xmlNewProp(node1, BAD_CAST "size_midi_cc", BAD_CAST buf);

	xmlSaveFormatFileEnc(filename, doc, "UTF-8", 1);

	xmlFreeDoc(doc);

	xmlCleanupParser();
	
	return 0;
}

