/***************************************************************************
 *            tranches.h
 *
 *  Copyright  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


 #ifndef tranches_h
 #define tranches_h


#include <fstream>
#include <sstream>
#include <iostream>
#include <cstdlib>
#include <vector>
#include <unistd.h>
#include <pthread.h>
#include <cstring>
#include <jack/jack.h>
#include <jack/ringbuffer.h>

#include "../midi/midiProcess.h"
#include "../audio/jackProcess.h"

#ifdef WITH_GUI
	#include "../ui/execWindow.h"
#endif

#ifdef WITH_LASH
	#include "../lash/lashClient.h"
#endif

#define MIDI_PRIORITY 60

#define TRANCHES_VERSION "0.1.0"

#define NB_MIDI_CONTROLS 5

class tranches  
{	
	public:
		tranches(int,char**); 
		~tranches();
		void showMessage(bool,std::string);
		std::string getFileName();
		std::string getCompleteFileName();

		int load(char*); 
		int start();
		void save(char*);
		void stop();

		int getOutputNb();
		void addOutput(std::string);
		void addCreateOutput(std::string);
		void removeOutput(int);
		int getOutputByName(std::string);
		std::string getOutputName(int);	
		void setOutputName(int,std::string);
		int getInputNb();
		void addInput(std::string,short,short);
		void addCreateInput(std::string,short,short);
		void removeInput(int);
		int getInputByName(std::string);
		std::string getInputName(int);	
		void setInputName(int,std::string);

		void setMidiControl(int,int,short);	
		short getMidiControl(int,int);
		void startMidiLearn();
		void stopMidiLearn();
		void processMidiLearn(int,int);
		int processMidiCC(unsigned short,unsigned short,unsigned short);	
		
		void repeat(int,bool);
		void reverse(int);
		void setSlice(int,bool);
		int getSlice(int);
		void setOutput(int);
		void setInput(int,bool);

		jackProcess* getJackProcess();			
		void updateJackStereoChannels();

	private:
		std::string fileName;
		bool loop;

		#ifdef WITH_GUI
			//Gui
			execWindow *execWin;
		#endif

		//Audio		
		jackProcess *jack;
		std::vector<std::string> jackOutputs;
		std::vector<std::string> jackInputs;
		std::vector<short> inputsMidiChannels;
		std::vector<short> inputsMidiCCs;

		std::vector< std::vector<int> > slicesArrays;
		int currentSlicesArray;
		
		//Midi
		midiProcess* midi;
	
		/*	midiControls Array
			
			2nd dim : 0 = channel , 1 = cc or note	

			first dim : 
			0 = input
			1 = output
			2 = sliceSize
			3 = swap
			4 = swapSlice1
			5 = swapSlice2
			6 = repeat
			7 = reverse
			8 = granulize
			9 = grainSize
		*/
		unsigned short midiControls[NB_MIDI_CONTROLS][2];


		#ifdef WITH_LASH
			//Lash
			lashClient *lashC;
		#endif
};

#endif
