/***************************************************************************
 *            midiProcess.cpp
 *
 *  Copyright  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "../base/tranches.h"
#include "midiProcess.h"
#include <iostream>
#include <cstdlib>
#include <vector>

using namespace std;

midiProcess::midiProcess(tranches *t,jack_client_t *j,int p):tra(t),jackClient(j),midiPriority(p),midiLearn(false)
{	
}

midiProcess::~midiProcess()
{
}

int midiProcess::midiInit()
{
	//Open the ALSA midi sequencer handler
  	if(snd_seq_open(&seqHandle, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) 
	{
    		return(-1);
  	}
	
  	//Client's name
  	snd_seq_set_client_name(seqHandle, "Tranches");
	
	if((midiInputPort = snd_seq_create_simple_port(seqHandle, "Tranches_IN",
              SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
              SND_SEQ_PORT_TYPE_APPLICATION)) < 0) 
	{
		return(-1);  
	}	

	return 0;
}

void midiProcess::startMidiLearn()
{
	midiLearn=true;
}

void midiProcess::stopMidiLearn()
{
	midiLearn=false;
}

void midiProcess::midiLoop()
{
		
	int npfd;
  	struct pollfd *pfd;
	npfd = snd_seq_poll_descriptors_count(seqHandle, POLLIN);
  	pfd = (struct pollfd *)alloca(npfd * sizeof(struct pollfd));
  	snd_seq_poll_descriptors(seqHandle, pfd, npfd, POLLIN);
  	while (1) 
	{
    		if (poll(pfd, npfd, 100000) > 0)
		{		
      			snd_seq_event_t *ev;

  			do 
			{
    				snd_seq_event_input(seqHandle, &ev); //we get the event from the seqHandler
    				snd_seq_ev_set_subs(ev);  
    				snd_seq_ev_set_direct(ev);	

				//if learning midi control
				if(midiLearn)
				{
					if (ev->type == SND_SEQ_EVENT_NOTEON)	
					{
						midiLearn=false;
						tra->processMidiLearn(ev->data.note.channel+1,ev->data.note.note);
					}
					else if(ev->type == SND_SEQ_EVENT_CONTROLLER)
					{
						midiLearn=false;
						tra->processMidiLearn(ev->data.control.channel+1,ev->data.control.param);
					}
				}

    				if (ev->type==SND_SEQ_EVENT_NOTEOFF || (ev->type==SND_SEQ_EVENT_NOTEON && ev->data.note.velocity==0))	
					tra->processMidiCC(ev->data.note.channel+1,ev->data.note.note,0);
				else
				if (ev->type == SND_SEQ_EVENT_NOTEON)
					tra->processMidiCC(ev->data.note.channel+1,ev->data.note.note,1);
				else		
				if (ev->type == SND_SEQ_EVENT_CONTROLLER)
				{
					tra->processMidiCC(ev->data.control.channel+1,ev->data.control.param,ev->data.control.value);		
				}
    				snd_seq_free_event(ev);
  			} while (snd_seq_event_input_pending(seqHandle, 0) > 0);
		}
	}

}




void* midiProcess::threadExec(void* pthis)	
{
	midiProcess* pparent = (midiProcess*) pthis; 	//gets a pointer to the current object
	pparent->midiLoop();			     	//starts the midi process loop
	pthread_exit(0);		
}	
	

int midiProcess::startThread()			
{
	int res =jack_client_create_thread(jackClient,&midiThread,midiPriority,true,threadExec,this);  	
	if(res!=0)
	{
		return -1;
	}
	
	
	return 0;
	  	
}

void midiProcess::stopThread()
{
	pthread_cancel(midiThread);	//stops the midiThread
	snd_seq_close(seqHandle);
}
	
int midiProcess::getAlsaID()
{
	return snd_seq_client_id(seqHandle);
}


