/***************************************************************************
 *            execWindow.cpp
 *
 *  Copyright  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "execWindow.h"
#include "../base/tranches.h"
#include <iostream>
#include <math.h>

using namespace std;


execWindow::execWindow(int w,int h,const char* titre,tranches *t):Fl_Window(w,h,titre),tra(t),fileSaved(true)
{
	//fltk init
	Fl::visual(FL_DOUBLE|FL_RGB);
	
	//window class
	this->xclass("tranches"); 

	//better look
	Fl::scheme("gtk+");

	//default theme
	Fl::set_color(FL_BACKGROUND_COLOR,240,240,240);
	Fl::set_color(FL_BACKGROUND2_COLOR,250,250,250);
	Fl::set_color(FL_FOREGROUND_COLOR,0,0,0);
	Fl::set_color(FL_INACTIVE_COLOR,200,200,200);
	Fl::set_color(FL_SELECTION_COLOR,100,100,100);

	//callback
	this->callback(statWindow,this);

	//menu
	Fl_Menu_Item menuitems[] = 
	{
  		{ "&Menu",0, 0, 0, FL_SUBMENU },
    			{ "&Open File",FL_CTRL + 'o',(Fl_Callback *)statOpen,this },
    			{ "&Save File",FL_CTRL + 's',(Fl_Callback *)statSave,this},
    			{ "S&ave File as",FL_CTRL + FL_SHIFT + 's',(Fl_Callback *)statSaveAs,this},
    			//{ "O&ptions", FL_CTRL + 'p',NULL,0},
    			{ "&Messages", FL_CTRL + 'm', statMessages,this},
    			{ "&Quit",FL_CTRL + 'q', (Fl_Callback *)statQuit,this},
			{0},
		{0}
	};
	menu = new Fl_Menu_Bar(0, 0, 400, 30);
	menu->copy(menuitems);

	//status
	status = new Fl_Group(0,h-15,w,15,"");			
	status->align(FL_ALIGN_INSIDE|FL_ALIGN_LEFT);

	//messages
	messages = new messageWindow(300,150,"Messages");

	//create the tabs
	tabs = new Fl_Tabs(0,30,w,h-50);
	tabs->when(FL_WHEN_CHANGED);
	tabs->callback(statTabs,this);
	execTab = new Fl_Group(0,50,w,h,"Execution");
	setTab = new Fl_Group(0,50,w,h,"Settings");
	tabs->insert(*setTab,0);
	tabs->insert(*execTab,1);
	this->insert(*tabs,0);
	
	//settings tab
	setInList = new Fl_Hold_Browser(5,70,140,70,"Stereo Inputs");
	setInList->align(FL_ALIGN_TOP|FL_ALIGN_CENTER);
	setInList->callback(statSetInList,this);
	setInName = new Fl_Input(15,150,70,20,"");
	setInName->align(FL_ALIGN_LEFT);
	setInName->callback(statSetInName,this);
	setInName->when(FL_WHEN_CHANGED);
	setInNew = new Fl_Button(90,150,20,20,"@+");
	setInNew->callback(statSetInNew,this);
	setInRemove = new Fl_Button(115,150,20,20,"@line");
	setInRemove->callback(statSetInRemove,this);
	setOutList = new Fl_Hold_Browser(150,70,140,70,"Stereo Outputs");
	setOutList->align(FL_ALIGN_TOP|FL_ALIGN_CENTER);
	setOutList->callback(statSetOutList,this);
	setOutName = new Fl_Input(160,150,70,20,"");
	setOutName->align(FL_ALIGN_LEFT);
	setOutName->callback(statSetOutName,this);
	setOutName->when(FL_WHEN_CHANGED);
	setOutNew = new Fl_Button(235,150,20,20,"@+");
	setOutNew->callback(statSetOutNew,this);
	setOutRemove = new Fl_Button(260,150,20,20,"@line");
	setOutRemove->callback(statSetOutRemove,this);
	setMidiSel = new Fl_Choice(50,h-135,200,20,"Midi Control");
	setMidiSel->align(FL_ALIGN_TOP|FL_ALIGN_CENTER);
	setMidiSel->callback(statSetMidiSel,this);
	setMidiAct = new Fl_Check_Button(20,h-90,20,20,"Active");
	setMidiAct->align(FL_ALIGN_TOP|FL_ALIGN_CENTER);
	setMidiAct->callback(statSetMidiAct,this);
	setMidiChan = new Fl_Counter(70,h-90,50,20,"Channel");
	setMidiChan->type(FL_SIMPLE_COUNTER);
	setMidiChan->align(FL_ALIGN_TOP);
	setMidiChan->step(1);
	setMidiChan->bounds(1,16);
	setMidiChan->callback(statSetMidiCC,this);
	setMidiChan->deactivate();
	setMidiCC = new Fl_Counter(140,h-90,50,20,"Control");
	setMidiCC->type(FL_SIMPLE_COUNTER);
	setMidiCC->align(FL_ALIGN_TOP);
	setMidiCC->step(1);
	setMidiCC->bounds(0,127);
	setMidiCC->callback(statSetMidiCC,this);
	setMidiCC->deactivate();
	setMidiLearn = new Fl_Button(210,h-90,50,20,"Learn");
	setMidiLearn->type(FL_TOGGLE_BUTTON);
	setMidiLearn->callback(statSetMidiLearn,this);
	setMidiLearn->deactivate();
	setTab->insert(*setInList,0);
	setTab->insert(*setInNew,1);
	setTab->insert(*setInRemove,2);
	setTab->insert(*setInName,3);
	setTab->insert(*setOutList,4);
	setTab->insert(*setOutNew,5);
	setTab->insert(*setOutRemove,6);
	setTab->insert(*setOutName,7);
	setTab->insert(*setMidiSel,8);
	setTab->insert(*setMidiAct,9);
	setTab->insert(*setMidiCC,10);
	setTab->insert(*setMidiChan,11);
	setTab->insert(*setMidiLearn,12);
	
	//execution tab
	//input and output
	execMixGroup = new Fl_Group(5,60,290,105);
	execMixGroup->box(FL_DOWN_BOX);
	//execIn = new Fl_Check_Browser(10,80,100,80,"Choose Inputs");
	execIn = new Fl_Choice(10,80,100,20,"Choose Input");
	execIn->callback(statExecIn,this);
	execIn->align(FL_ALIGN_TOP|FL_ALIGN_LEFT);
	execOut = new Fl_Choice(150,80,100,20,"Choose Output");
	execOut->callback(statExecOut,this);
	execOut->align(FL_ALIGN_TOP|FL_ALIGN_LEFT);
	execMixGroup->insert(*execIn,0);
	execMixGroup->insert(*execOut,1);

	//slice
	execSliceGroup = new Fl_Group(5,175,290,30);
	execSliceGroup->box(FL_BORDER_BOX);
	execSliceDis = new Fl_Output(180,180,70,20,"");
	execSlice = new Fl_Slider(80,180,100,20,"Slice:");	
	execSlice->align(FL_ALIGN_LEFT);
	execSlice->type(FL_HOR_NICE_SLIDER);
	execSlice->bounds(0,7);
	execSlice->callback(statExecSlice,this);
	execSlice->value(4);
	execSliceDis->value("1 beat");
	execSliceGroup->insert(*execSliceDis,0);
	execSliceGroup->insert(*execSlice,1);

/*
	//swap	
	execSwapGroup = new Fl_Group(5,215,290,30);
	execSwapGroup->box(FL_UP_BOX);
	execSwap = new Fl_Button(15,220,50,20,"Swap");
	execSwap->when(FL_WHEN_CHANGED);
	execSwap->callback(statExecSwap,this);
	execSwapS1 = new Fl_Value_Slider(75,220,100,20,"");	
	execSwapS1->type(FL_HOR_SLIDER);
	execSwapS1->bounds(1,16);
	execSwapS1->callback(statExecSwapS1,this);
	execSwapS1->value(1);
	execSwapS1->step(1);
	execSwapS2 = new Fl_Value_Slider(190,220,100,20,"&");	
	execSwapS2->type(FL_HOR_SLIDER);
	execSwapS2->align(FL_ALIGN_LEFT);
	execSwapS2->bounds(1,16);
	execSwapS2->callback(statExecSwapS2,this);
	execSwapS2->value(3);
	execSwapS2->step(1);
	execSwapGroup->insert(*execSwap,0);
	execSwapGroup->insert(*execSwapS1,1);
	execSwapGroup->insert(*execSwapS2,2);
*/
	//repeat
	execRepeatGroup = new Fl_Group(5,255,290,30);
	execRepeatGroup->box(FL_UP_BOX);
	execRepeat = new Fl_Toggle_Button(15,260,60,20,"Repeat");
	execRepeat->when(FL_WHEN_CHANGED);
	execRepeat->callback(statExecRepeat,this);
	execReverse = new Fl_Toggle_Button(100,262,60,15,"Reverse");
	execReverse->callback(statExecReverse,this);
	execRepeatGroup->insert(*execRepeat,0);
	execRepeatGroup->insert(*execReverse,1);
/*		
	//granulize	
	execGranGroup = new Fl_Group(5,295,290,30);
	execGranGroup->box(FL_UP_BOX);
	execGran = new Fl_Button(15,300,70,20,"Granulize");
	execGran->when(FL_WHEN_CHANGED);
	execGran->callback(statExecGran,this);
	execGranSize = new Fl_Value_Slider(140,300,100,20,"Size: ");	
	execGranSize->align(FL_ALIGN_LEFT);
	execGranSize->type(FL_HOR_SLIDER);
	execGranSize->bounds(16,128);
	execGranSize->step(1);
	execGranSize->callback(statExecGranSize,this);
	execGranSize->value(16);
	execGranGroup->insert(*execGran,0);
	execGranGroup->insert(*execGranSize,1);
*/
	execTab->insert(*execMixGroup,0);
	execTab->insert(*execSliceGroup,1);
//	execTab->insert(*execSwapGroup,2);
	execTab->insert(*execRepeatGroup,3);
//	execTab->insert(*execGranGroup,4);

}

execWindow::~execWindow()
{
}

void execWindow::cbWindow(Fl_Widget*)
{
	if (Fl::event()==FL_SHORTCUT && Fl::event_key()==FL_Escape) 
    		return; // ignore Escape
	
	cbQuit(this);
	exit(0);	
}

void execWindow::reset()
{	
	//clear the messages
//	messages->clear();

	//clear the settings tab
	setInList->clear();
	setOutList->clear();
	setMidiSel->clear();
	
	//clear the exec tab
	execIn->clear();
	execOut->clear();
	execSlice->value(5);
	execSliceDis->value("1 beat");
	//execSwapS1->value(1);
	//execSwapS2->value(3);
	execReverse->value(0);
	//execGranSize->value(16);

	this->redraw();
	Fl::check();
	
}

void execWindow::init()
{
	//we set the midi controls
	setMidiSel->add("Select Input");
	setMidiSel->add("Select Output");
	setMidiSel->add("Select Slice Size");
//	setMidiSel->add("Trigger Swap");
//	setMidiSel->add("Select Swap Slice 1");
//	setMidiSel->add("Select Swap Slice 2");
	setMidiSel->add("Trigger Repeat");
	setMidiSel->add("Toggle Reverse");
//	setMidiSel->add("Trigger Granulize");
//	setMidiSel->add("Select Grain Size");

	//we set the values of the inputs
	execIn->add("All inputs");
	for(int i=0;i<tra->getInputNb();++i)
	{
		std::string inName = tra->getInputName(i);
		execIn->add(inName.c_str());
		setInList->add(inName.c_str());
		//inName = "Toggle Input "+inName;
		//setMidiSel->add(inName.c_str());
	}

	//we set the values of the outputs
	for(int i=0;i<tra->getOutputNb();++i)
	{
		execOut->add((tra->getOutputName(i)).c_str());
		setOutList->add((tra->getOutputName(i)).c_str());
	}
	execOut->value(0);

	tabs->value(execTab);

	this->redraw();
	Fl::check();
	
}

//-----------------------------------MENU--------------------------------
void execWindow::cbOpen(Fl_Widget* w)
{
	//open file dialog
	tardiFlFileBrowser *tFB = new tardiFlFileBrowser();
	tFB->open("Open .tra file","{*.tra}",tra->getCompleteFileName());
	while(tFB->shown())
	{
		Fl::wait();
	}
	char* newFile=tFB->getResult();
	if(newFile)
	{
		this->reset();
		tra->load(newFile);
		tra->start();
		delete [] newFile;
	}
	delete tFB;
}

void execWindow::cbSave(Fl_Widget* w)
{
	if(tra->getCompleteFileName()=="")
	{
		cbSaveAs(this);
	}	
	else
	{
		size_t size = tra->getCompleteFileName().size() + 1;
		char * buffer = new char[ size ];
		strncpy( buffer, tra->getCompleteFileName().c_str(), size );
		tra->save(buffer);
		delete [] buffer;
		std::string m = "Saved as "+tra->getFileName();
		fileSaved=true;
		setStatus(m);
	}
}

void execWindow::cbSaveAs(Fl_Widget* w)
{
	//save file dialog
	tardiFlFileBrowser *tFB = new tardiFlFileBrowser();
	tFB->open("Save .tra file ","{*.tra}",tra->getCompleteFileName());
	while(tFB->shown())
	{
		Fl::wait();
	}
	char* newFile=tFB->getResult();
	if (newFile)
	{
		tra->save(newFile);
		std::string f = tra->getFileName();
		std::string m = "Saved as "+f;
		setStatus(m);
		fileSaved=true;
		delete [] newFile;
	}
	delete tFB;
}

void execWindow::cbQuit(Fl_Widget* w)
{
	if(!fileSaved)
	{
		switch(fl_choice("File not saved ! Quit without saving ?", "Quit without saving","Save then Quit" ,"Don't quit"))
		{
			case 0:
			break;
			case 1:
			{
				cbSave(this);
				if(!fileSaved)
					return;
			}
			break;
			default: return;
				
		}
	}	

	//tout nettoyer puis quitter
	this->reset();
	tra->stop();

	exit(0);
}

void execWindow::setTitle(std::string t)
{
	label(t.c_str());	
}

//--------------------------------MESSAGES-------------------------------------
void execWindow::cbMessages(Fl_Widget*)
{
	messages->show();
}

void execWindow::showMessage(bool e,std::string mess)
{
	setStatus(mess);
	messages->addMessage(mess);
	if(e)
	{
		messages->show();
	}
}

void execWindow::setStatus(std::string st)
{
	statusStr = st;
	status->label(statusStr.c_str());
	status->redraw();
	Fl::check();
}

//---------------------------------TABS-----------------------------------------
void execWindow::cbTabs(Fl_Widget*)
{
	setStatus("");
}

//------------------------------SETTINGS-------------------------------------
void execWindow::cbSetMidiSel(Fl_Widget*)
{
	fileSaved=false;
	short chan = tra->getMidiControl(setMidiSel->value(),0);	
	short cc = tra->getMidiControl(setMidiSel->value(),1);	
	setMidiChan->value(chan);
	setMidiCC->value(cc);
	if(chan>=0)
	{
		setMidiAct->value(1);
		cbSetMidiAct(this);
	}	
	else
	{
		setMidiAct->value(0);
		cbSetMidiAct(this);
	}
}

void execWindow::cbSetMidiAct(Fl_Widget*)
{
	fileSaved=false;
	if(setMidiAct->value())
	{
		setMidiCC->activate();
		setMidiChan->activate();
		setMidiLearn->activate();
	}
	else
	{
		setMidiCC->deactivate();
		setMidiChan->deactivate();
		setMidiLearn->deactivate();
		setMidiCC->value(-1);
		setMidiChan->value(-1);
	}

	updateMidi();

	this->redraw();
	Fl::check();
}

void execWindow::cbSetMidiCC(Fl_Widget*)
{
	updateMidi();
}

void execWindow::cbSetMidiChan(Fl_Widget*)
{
	updateMidi();
}

void execWindow::cbSetMidiLearn(Fl_Widget*)
{
	if(setMidiLearn->value())
	{
		setStatus("Midi Learn: Tweak the controller");
		tra->startMidiLearn();
	}
	else
	{
		setStatus("Midi Learn: Stopped");
		tra->stopMidiLearn();
	}
		
}

void execWindow::learnMidi(int ch,int ccn)
{
	setMidiLearn->value(0);
	setMidiCC->value(ccn);
	setMidiChan->value(ch);
	setMidiCC->redraw();
	setMidiChan->redraw();
	setStatus("Midi Learn: Done");
	updateMidi();
}

void execWindow::updateMidi()
{
	fileSaved=false;
	tra->setMidiControl(setMidiSel->value(),0,(short)(setMidiChan->value()));
	tra->setMidiControl(setMidiSel->value(),1,(short)(setMidiCC->value()));
}


void execWindow::cbSetOutList(Fl_Widget*)
{
	fileSaved=false;
	if(setOutList->value()>0)
	{
		setOutName->value(setOutList->text(setOutList->value()));
	}
	else
		setOutName->value("");
}

void execWindow::cbSetOutNew(Fl_Widget*)
{
	fileSaved=false;
	fileSaved=false;
	bool outputCreated=false;
	unsigned int outputNb=0;
	std::string outputName; 
	while(!outputCreated)
	{
		std::ostringstream oss;
		oss << outputNb;	
		outputName ="Output"+oss.str(); 
		outputCreated=true;
		for(unsigned int i=0;i<tra->getOutputNb();++i)
		{
			if(tra->getOutputName(i)==outputName)
			{
				outputCreated=false;
			}
		}
		++outputNb;
	}
	tra->addCreateOutput(outputName);
	setOutList->add(outputName.c_str());
	execOut->add(outputName.c_str());
}

void execWindow::cbSetOutRemove(Fl_Widget*)
{
	fileSaved=false;
	//remove the output, if there is one left after that
	if(tra->getOutputNb()>1 && setOutList->value()>0)
	{
		tra->removeOutput(setOutList->value()-1);	
		execOut->remove(setOutList->value()-1);
		setOutList->remove(setOutList->value());
	}
}

void execWindow::cbSetOutName(Fl_Widget*)
{
	fileSaved=false;
	if(setOutList->value()>0)
	{
		tra->setOutputName(setOutList->value()-1,std::string(setOutName->value()));
		setOutList->text(setOutList->value(),setOutName->value());
		execOut->replace(setOutList->value()-1,setOutName->value());
	}
}

void execWindow::cbSetInList(Fl_Widget*)
{
	fileSaved=false;
	if(setInList->value()>0)
	{
		setInName->value(setInList->text(setInList->value()));
	}
	else
		setInName->value("");
}

void execWindow::cbSetInNew(Fl_Widget*)
{
	fileSaved=false;
	bool inputCreated=false;
	unsigned int inputNb=0;
	std::string inputName; 
	while(!inputCreated)
	{
		std::ostringstream oss;
		oss << inputNb;	
		inputName ="Input"+oss.str(); 
		inputCreated=true;
		for(unsigned int i=0;i<tra->getInputNb();++i)
		{
			if(tra->getInputName(i)==inputName)
			{
				inputCreated=false;
			}
		}
		++inputNb;
	}
	tra->addCreateInput(inputName,-1,-1);
	setInList->add(inputName.c_str());
	execIn->add(inputName.c_str());
}

void execWindow::cbSetInRemove(Fl_Widget*)
{
	fileSaved=false;
	//remove the input, if there is one left after that
	if(tra->getInputNb()>1 && setInList->value()>0)
	{
		tra->removeInput(setInList->value()-1);	
		execIn->remove(setInList->value());
		setInList->remove(setInList->value());
	}
}

void execWindow::cbSetInName(Fl_Widget*)
{
	fileSaved=false;
	if(setInList->value()>0)
	{
		tra->setInputName(setInList->value()-1,std::string(setInName->value()));
		setInList->text(setInList->value(),setInName->value());
		execIn->clear();
		execIn->add("All Inputs");
		//we set the values of the inputs
		for(int i=0;i<tra->getInputNb();++i)
		{
			std::string inName = tra->getInputName(i);
			execIn->add(inName.c_str());
		}
	}
}

//------------------------------EXECUTION-------------------------------------

void execWindow::cbExecIn(Fl_Widget*)
{
	tra->setInput(execIn->value()-1,true);	
}

void execWindow::changeInput(int in)
{
	execIn->value(in+1);
	execIn->redraw();
	Fl::check();
}

void execWindow::cbExecOut(Fl_Widget*)
{
	tra->setOutput(execOut->value());	
}

void execWindow::changeOutput(int out)
{
	execOut->value(out);
	execOut->redraw();
	Fl::check();

}

void execWindow::cbExecSlice(Fl_Widget*)
{
	tra->setSlice((int)(execSlice->value()),false);	
}

void execWindow::changeSlice(int val,int slice)
{
	std::string sl; 
	std::ostringstream oss;

	if(slice<0)
	{
		oss << (-slice);	
		sl="1/"+oss.str()+" beat";
	}
	else
	{
		oss << slice;
		if(slice==1)
			sl=oss.str()+" beat";
		else
			sl=oss.str()+" beats";
			
	}
	execSliceDis->value(sl.c_str());
	execSlice->value(val);
	execSliceDis->redraw();
	execSlice->redraw();
	Fl::check();
}

void execWindow::cbExecSwap(Fl_Widget*)
{
}

void execWindow::cbExecSwapS1(Fl_Widget*)
{
}

void execWindow::cbExecSwapS2(Fl_Widget*)
{
}

void execWindow::cbExecRepeat(Fl_Widget*)
{
	tra->repeat(execRepeat->value(),true);
}


void execWindow::cbExecReverse(Fl_Widget*)
{
	tra->reverse(execReverse->value());
}

void execWindow::toggleExecRepeat(int val)
{
	execRepeat->value(val);
	execRepeat->redraw();
	Fl::check();
}

void execWindow::toggleExecReverse(int val)
{
	execReverse->value(val);
	execReverse->redraw();
	Fl::check();
}

void execWindow::cbExecGran(Fl_Widget*)
{
}

void execWindow::cbExecGranSize(Fl_Widget*)
{
}



